import numpy as np
import pandas as pd
import scipy.linalg
from scipy.optimize import minimize


class FisherMatrixPosteriorEstimator(object):
    def __init__(self, likelihood, priors, parameters=None, fd_eps=1e-6, n_prior_samples=100):
        """ A class to estimate posteriors using the Fisher Matrix approach

        Parameters
        ----------
        likelihood: bilby.core.likelihood.Likelihood
            A bilby likelihood object
        priors: bilby.core.prior.PriorDict
            A bilby prior object
        parameters: list
            Names of parameters to sample in
        fd_eps: float
            A parameter to control the size of perturbation used when finite
            differencing the likelihood
        n_prior_samples: int
            The number of prior samples to draw and use to attempt estimatation
            of the maximum likelihood sample.
        """
        self.likelihood = likelihood
        if parameters is None:
            self.parameter_names = priors.non_fixed_keys
        else:
            self.parameter_names = parameters
        self.fd_eps = fd_eps
        self.n_prior_samples = n_prior_samples
        self.N = len(self.parameter_names)

        self.prior_samples = [
            priors.sample_subset(self.parameter_names) for _ in range(n_prior_samples)
        ]
        self.prior_bounds = [(priors[key].minimum, priors[key].maximum) for key in self.parameter_names]

        self.prior_width_dict = {}
        for key in self.parameter_names:
            width = priors[key].width
            if np.isnan(width):
                raise ValueError(f"Prior width is ill-formed for {key}")
            self.prior_width_dict[key] = width

    def log_likelihood(self, sample):
        self.likelihood.parameters.update(sample)
        return self.likelihood.log_likelihood()

    def calculate_iFIM(self, sample):
        FIM = self.calculate_FIM(sample)
        iFIM = scipy.linalg.inv(FIM)

        # Ensure iFIM is positive definite
        min_eig = np.min(np.real(np.linalg.eigvals(iFIM)))
        if min_eig < 0:
            iFIM -= 10 * min_eig * np.eye(*iFIM.shape)

        return iFIM

    def sample_array(self, sample, n=1):
        from .utils.random import rng

        if sample == "maxL":
            sample = self.get_maximum_likelihood_sample()

        self.mean = np.array(list(sample.values()))
        self.iFIM = self.calculate_iFIM(sample)
        return rng.multivariate_normal(self.mean, self.iFIM, n)

    def sample_dataframe(self, sample, n=1):
        samples = self.sample_array(sample, n)
        return pd.DataFrame(samples, columns=self.parameter_names)

    def calculate_FIM(self, sample):
        FIM = np.zeros((self.N, self.N))
        for ii, ii_key in enumerate(self.parameter_names):
            for jj, jj_key in enumerate(self.parameter_names):
                FIM[ii, jj] = -self.get_second_order_derivative(sample, ii_key, jj_key)

        return FIM

    def get_second_order_derivative(self, sample, ii, jj):
        if ii == jj:
            return self.get_finite_difference_xx(sample, ii)
        else:
            return self.get_finite_difference_xy(sample, ii, jj)

    def get_finite_difference_xx(self, sample, ii):
        # Sample grid
        p = self.shift_sample_x(sample, ii, 1)
        m = self.shift_sample_x(sample, ii, -1)

        dx = .5 * (p[ii] - m[ii])

        loglp = self.log_likelihood(p)
        logl = self.log_likelihood(sample)
        loglm = self.log_likelihood(m)

        return (loglp - 2 * logl + loglm) / dx ** 2

    def get_finite_difference_xy(self, sample, ii, jj):
        # Sample grid
        pp = self.shift_sample_xy(sample, ii, 1, jj, 1)
        pm = self.shift_sample_xy(sample, ii, 1, jj, -1)
        mp = self.shift_sample_xy(sample, ii, -1, jj, 1)
        mm = self.shift_sample_xy(sample, ii, -1, jj, -1)

        dx = .5 * (pp[ii] - mm[ii])
        dy = .5 * (pp[jj] - mm[jj])

        loglpp = self.log_likelihood(pp)
        loglpm = self.log_likelihood(pm)
        loglmp = self.log_likelihood(mp)
        loglmm = self.log_likelihood(mm)

        return (loglpp - loglpm - loglmp + loglmm) / (4 * dx * dy)

    def shift_sample_x(self, sample, x_key, x_coef):

        vx = sample[x_key]
        dvx = self.fd_eps * self.prior_width_dict[x_key]

        shift_sample = sample.copy()
        shift_sample[x_key] = vx + x_coef * dvx

        return shift_sample

    def shift_sample_xy(self, sample, x_key, x_coef, y_key, y_coef):

        vx = sample[x_key]
        vy = sample[y_key]

        dvx = self.fd_eps * self.prior_width_dict[x_key]
        dvy = self.fd_eps * self.prior_width_dict[y_key]

        shift_sample = sample.copy()
        shift_sample[x_key] = vx + x_coef * dvx
        shift_sample[y_key] = vy + y_coef * dvy
        return shift_sample

    def get_maximum_likelihood_sample(self, initial_sample=None):
        """ A method to attempt optimization of the maximum likelihood

        This uses a simple scipy optimization approach, starting from a number
        of draws from the prior to avoid problems with local optimization.

        Note: this approach works well in small numbers of dimensions when the
        posterior is narrow relative to the prior. But, if the number of dimensions
        is large or the posterior is wide relative to the prior, the method fails
        to find the global maximum in high dimensional problems.
        """
        minlogL = np.inf
        for i in range(self.n_prior_samples):
            initial_sample = self.prior_samples[i]

            x0 = list(initial_sample.values())

            def neg_log_like(x, self, T=1):
                sample = {key: val for key, val in zip(self.parameter_names, x)}
                return - 1 / T * self.log_likelihood(sample)

            out = minimize(
                neg_log_like,
                x0,
                args=(self, 1),
                bounds=self.prior_bounds,
                method="L-BFGS-B",
            )
            if out.fun < minlogL:
                minout = out

        return {key: val for key, val in zip(self.parameter_names, minout.x)}



#!/usr/bin/env python
"""
Tutorial to demonstrate running parameter estimation on a binary neutron star
with equation of state inference using a 3 piece dynamic polytrope model.

This example estimates the masses and polytrope model parameters.

This polytrope model is generically introduced in https://arxiv.org/pdf/1804.04072.pdf.
LALSimulation has an implementation of this generic model with 3 polytrope pieces attached
to a crust EOS, and that can be found in LALSimNeutronStarEOSDynamicPolytrope.c.

Details on a 4 piece static polytrope model can be found here https://arxiv.org/pdf/0812.2163.pdf.
The reparameterization here uses a 3-piece dynamic polytrope model implemented in lalsimulation
LALSimNeutronStarEOSDynamicPolytrope.c.
"""


import bilby
import numpy as np

# Specify the output directory and the name of the simulation.
outdir = "outdir"
label = "bns_polytrope_eos_simulation"
bilby.core.utils.setup_logger(outdir=outdir, label=label)
# Set up a random seed for result reproducibility.  This is optional!
np.random.seed(88170235)

# We are going to inject a binary neutron star waveform with component mass and lambda
# values consistent with the MPA1 equation of state.
# The lambda values are generated from LALSimNeutronStarEOS_MPA1.dat
# lalsim-ns-params -n MPA1 -m 1.523194
# lalsim-ns-params -n MPA1 -m 1.5221469
# Note that we injection with tidal params (lambdas) but recover in eos model params.
injection_parameters = dict(
    mass_1=1.523194,
    mass_2=1.5221469,
    lambda_1=311.418,
    lambda_2=312.717,
    chi_1=0.0001,
    chi_2=0.0001,
    luminosity_distance=57.628867,
    theta_jn=0.66246341,
    psi=0.18407784,
    phase=5.4800181,
    geocent_time=10.0,
    ra=3.6309322,
    dec=-0.30355747,
)

# Set the duration and sampling frequency of the data segment that we're going
# to inject the signal into. For the
# TaylorF2 waveform, we cut the signal close to the isco frequency
duration = 64
sampling_frequency = 2 * 2048
start_time = injection_parameters["geocent_time"] + 2 - duration

# Fixed arguments passed into the source model. The analysis starts at 40 Hz.
waveform_arguments = dict(
    waveform_approximant="TaylorF2",
    reference_frequency=30.0,
    minimum_frequency=30.0,
)

# Create the waveform_generator using a LAL Binary Neutron Star source function
waveform_generator = bilby.gw.WaveformGenerator(
    duration=duration,
    sampling_frequency=sampling_frequency,
    frequency_domain_source_model=bilby.gw.source.lal_binary_neutron_star,
    parameter_conversion=bilby.gw.conversion.convert_to_lal_binary_neutron_star_parameters,
    waveform_arguments=waveform_arguments,
)

# Set up interferometers.  In this case we'll use three interferometers
# (LIGO-Hanford (H1), LIGO-Livingston (L1), and Virgo (V1)).
# These default to their design sensitivity and start at 40 Hz.
interferometers = bilby.gw.detector.InterferometerList(["H1", "L1", "V1"])
for interferometer in interferometers:
    interferometer.minimum_frequency = 30.0
interferometers.set_strain_data_from_zero_noise(
    sampling_frequency=sampling_frequency, duration=duration, start_time=start_time
)
interferometers.inject_signal(
    parameters=injection_parameters, waveform_generator=waveform_generator
)

# Load the default prior for binary neutron stars.
# We're going to sample in chirp_mass, mass_ratio, and model parameters
# rather than mass_1, mass_2, or lamba_tilde, delta_lambda_tilde.
# BNS have aligned spins by default, if you want to allow precessing spins
# pass aligned_spin=False to the BNSPriorDict
priors = bilby.gw.prior.BNSPriorDict()

priors["chirp_mass"] = bilby.gw.prior.Uniform(0.4, 4.4)
priors["mass_ratio"] = bilby.gw.prior.Uniform(0.125, 1)

# The following are dynamic polytrope model priors
# They are required for EOS inference
priors["eos_polytrope_gamma_0"] = bilby.core.prior.Uniform(
    1.0, 5.0, name="Gamma0", latex_label="$\\Gamma_0$"
)
priors["eos_polytrope_gamma_1"] = bilby.core.prior.Uniform(
    1.0, 5.0, name="Gamma1", latex_label="$\\Gamma_1$"
)
priors["eos_polytrope_gamma_2"] = bilby.core.prior.Uniform(
    1.0, 5.0, name="Gamma2", latex_label="$\\Gamma_2$"
)
"""
One can run this model without the reparameterization using the following
priors in place of the scaled pressure priors. The reparameterization approximates
the flat priors in pressure however the reparameterization initializes twice as fast.

priors["eos_polytrope_log10_pressure_1"] = bilby.core.prior.Uniform(
    34.0, 37.0, name="plogp1", latex_label="$log(p_1)$"
)
priors["eos_polytrope_log10_pressure_2"] = bilby.core.prior.Uniform(
    34.0, 37.0, name="plogp2", latex_label="$log(p_2)$"
)
"""
priors["eos_polytrope_scaled_pressure_ratio"] = bilby.core.prior.Uniform(
    minimum=0.0, maximum=1, name="Pressure_Ratio", latex_label="$\\rho$"
)
priors["eos_polytrope_scaled_pressure_2"] = bilby.core.prior.Triangular(
    minimum=0.0, maximum=4.0, mode=4, name="Pressure_offset", latex_label="$p_0$"
)
priors["eos_check"] = bilby.gw.prior.EOSCheck()
# The eos_check sets up a rejection prior for sampling.
# It should be enabled for all eos runs.

# Pinning other parameters to their injected values.
for key in [
    "luminosity_distance",
    "geocent_time",
    "chi_1",
    "chi_2",
    "psi",
    "phase",
    "ra",
    "dec",
    "theta_jn",
]:
    priors[key] = injection_parameters[key]

# Remove tidal parameters from our priors.
priors.pop("lambda_1")
priors.pop("lambda_2")

# Initialise the likelihood by passing in the interferometer data (IFOs)
# and the waveform generator
likelihood = bilby.gw.GravitationalWaveTransient(
    interferometers=interferometers,
    waveform_generator=waveform_generator,
    time_marginalization=False,
    phase_marginalization=False,
    distance_marginalization=False,
    priors=priors,
)

# Run sampler.  This model has mostly been testing with the `dynesty` sampler.
result = bilby.run_sampler(
    conversion_function=bilby.gw.conversion.generate_all_bns_parameters,
    likelihood=likelihood,
    priors=priors,
    sampler="dynesty",
    nlive=1000,
    dlogz=0.1,
    injection_parameters=injection_parameters,
    outdir=outdir,
    label=label,
)

result.plot_corner()



import unittest
import os
from shutil import rmtree
from importlib.metadata import version

import numpy as np
import lal
import lalsimulation as lalsim
from gwpy.timeseries import TimeSeries
from gwpy.detector import Channel
from scipy.stats import ks_2samp
import pytest

import bilby
from bilby.gw import utils as gwutils


class TestGWUtils(unittest.TestCase):
    def setUp(self):
        self.outdir = "outdir"
        bilby.core.utils.check_directory_exists_and_if_not_mkdir(self.outdir)

    def tearDown(self):
        try:
            rmtree(self.outdir)
        except FileNotFoundError:
            pass

    def test_asd_from_freq_series(self):
        freq_data = np.array([1, 2, 3])
        df = 0.1
        asd = gwutils.asd_from_freq_series(freq_data, df)
        self.assertTrue(np.all(asd == freq_data * 2 * df ** 0.5))

    def test_psd_from_freq_series(self):
        freq_data = np.array([1, 2, 3])
        df = 0.1
        psd = gwutils.psd_from_freq_series(freq_data, df)
        self.assertTrue(np.all(psd == (freq_data * 2 * df ** 0.5) ** 2))

    def test_inner_product(self):
        aa = np.array([1, 2, 3])
        bb = np.array([5, 6, 7])
        frequency = np.array([0.2, 0.4, 0.6])
        PSD = bilby.gw.detector.PowerSpectralDensity.from_aligo()
        ip = gwutils.inner_product(aa, bb, frequency, PSD)
        self.assertEqual(ip, 0)

    def test_noise_weighted_inner_product(self):
        aa = np.array([1e-23, 2e-23, 3e-23])
        bb = np.array([5e-23, 6e-23, 7e-23])
        frequency = np.array([100, 101, 102])
        PSD = bilby.gw.detector.PowerSpectralDensity.from_aligo()
        psd = PSD.power_spectral_density_interpolated(frequency)
        duration = 4
        nwip = gwutils.noise_weighted_inner_product(aa, bb, psd, duration)
        self.assertEqual(nwip, 239.87768033598326)

        self.assertEqual(
            gwutils.optimal_snr_squared(aa, psd, duration),
            gwutils.noise_weighted_inner_product(aa, aa, psd, duration),
        )

    def test_matched_filter_snr(self):
        signal = np.array([1e-23, 2e-23, 3e-23])
        frequency_domain_strain = np.array([5e-23, 6e-23, 7e-23])
        frequency = np.array([100, 101, 102])
        PSD = bilby.gw.detector.PowerSpectralDensity.from_aligo()
        psd = PSD.power_spectral_density_interpolated(frequency)
        duration = 4

        mfsnr = gwutils.matched_filter_snr(
            signal, frequency_domain_strain, psd, duration
        )
        self.assertEqual(mfsnr, 25.510869054168282)

    @pytest.mark.skip(reason="GWOSC unstable: avoiding this test")
    def test_get_event_time(self):
        from urllib3.exceptions import NewConnectionError
        events = [
            "GW150914",
            "GW170104",
        ]
        for event in events:
            try:
                self.assertTrue(isinstance(gwutils.get_event_time(event), float))
            except NewConnectionError:
                return

        with self.assertRaises(ValueError):
            gwutils.get_event_time("GW010290")

    @pytest.mark.skipif(version("gwpy") < "3.0.8", reason="GWpy version < 3.0.8")
    def test_read_frame_file(self):
        """
        Test that reading a frame file works as expected
        for a few conditions.

        1. Reading without time limits returns the full data
        2. Reading with time limits returns the expected data
           (inclusive of start time if present, exclusive of end time)
        3. Reading without the channel name provided finds a standard name
        4. Reading without the channel with a non-standard name returns None.

        Notes
        =====
        There was a longstanding bug in gwpy that we previously tested for
        here, but this has been fixed in gwpy 3.0.8.
        """
        start_time = 0
        end_time = 10
        channel = "H1:GDS-CALIB_STRAIN"
        N = 100
        times = np.linspace(start_time, end_time, N, endpoint=False)
        data = np.random.normal(0, 1, N)
        ts = TimeSeries(data=data, times=times, t0=0)
        ts.channel = Channel(channel)
        ts.name = channel
        filename = os.path.join(self.outdir, "test.gwf")
        ts.write(filename, format="gwf")

        # Check reading without time limits
        strain = gwutils.read_frame_file(
            filename, start_time=None, end_time=None, channel=channel
        )
        self.assertEqual(strain.name, channel)
        self.assertTrue(np.all(strain.value == data))

        # Check reading with time limits
        start_cut = 2
        end_cut = 8
        strain = gwutils.read_frame_file(
            filename, start_time=start_cut, end_time=end_cut, channel=channel
        )
        idxs = (times >= start_cut) & (times < end_cut)
        self.assertTrue(np.all(strain.value == data[idxs]))

        # Check reading with unknown channels
        strain = gwutils.read_frame_file(filename, start_time=None, end_time=None)
        self.assertTrue(np.all(strain.value == data))

        # Check reading with incorrect channel
        strain = gwutils.read_frame_file(
            filename, start_time=None, end_time=None, channel="WRONG"
        )
        self.assertTrue(np.all(strain.value == data))

        ts = TimeSeries(data=data, times=times, t0=0)
        ts.name = "NOT-A-KNOWN-CHANNEL"
        ts.write(filename, format="gwf")
        strain = gwutils.read_frame_file(filename, start_time=None, end_time=None)
        self.assertEqual(strain, None)

    def test_convert_args_list_to_float(self):
        self.assertEqual(
            gwutils.convert_args_list_to_float(1, "2", 3.0), [1.0, 2.0, 3.0]
        )
        with self.assertRaises(ValueError):
            gwutils.convert_args_list_to_float(1, "2", "ten")

    def test_lalsim_SimInspiralTransformPrecessingNewInitialConditions(self):
        a = gwutils.lalsim_SimInspiralTransformPrecessingNewInitialConditions(
            0.1, 0, 0.6, 0.5, 0.6, 0.1, 0.8, 30.6, 23.2, 50, 0
        )
        self.assertTrue(len(a) == 7)

    def test_get_approximant(self):
        with self.assertRaises(ValueError):
            gwutils.lalsim_GetApproximantFromString(10)

    def test_lalsim_SimInspiralChooseFDWaveform(self):
        a = gwutils.lalsim_SimInspiralChooseFDWaveform(
            35.2,
            20.4,
            0.1,
            0.2,
            0.2,
            0.2,
            0.2,
            0.1,
            1000,
            2,
            2.3,
            45.0,
            0.1,
            10,
            0.01,
            10,
            1000,
            20,
            None,
            lalsim.IMRPhenomPv2,
        )
        self.assertEqual(len(a), 2)
        self.assertEqual(type(a[0]), lal.COMPLEX16FrequencySeries)
        self.assertEqual(type(a[1]), lal.COMPLEX16FrequencySeries)

        with self.assertRaises(RuntimeError):
            _ = gwutils.lalsim_SimInspiralChooseFDWaveform(
                35.2,
                20.4,
                0.1,
                0.2,
                0.2,
                0.2,
                0.2,
                0.1,
                1000,
                2,
                2.3,
                45.0,
                0.1,
                10,
                0.01,
                10,
                1000,
                20,
                None,
                "Fail",
            )

        with self.assertRaises(ValueError):
            _ = gwutils.lalsim_SimInspiralChooseFDWaveform(
                35.2,
                20.4,
                0.1,
                0.2,
                0.2,
                0.2,
                0.2,
                0.1,
                1000,
                2,
                2.3,
                45.0,
                0.1,
                10,
                0.01,
                10,
                1000,
                20,
                None,
                1.5,
            )


class TestSkyFrameConversion(unittest.TestCase):

    def setUp(self) -> None:
        self.priors = bilby.core.prior.PriorDict()
        self.priors["ra"] = bilby.core.prior.Uniform(0, 2 * np.pi)
        self.priors["dec"] = bilby.core.prior.Cosine()
        self.priors["azimuth"] = bilby.core.prior.Uniform(0, 2 * np.pi)
        self.priors["zenith"] = bilby.core.prior.Sine()
        self.priors["time"] = bilby.core.prior.Uniform(-0.1, 0.1)
        self.ifos = bilby.gw.detector.InterferometerList(["H1", "L1"])
        self.samples = self.priors.sample(10000)

    def tearDown(self) -> None:
        del self.priors
        del self.ifos
        del self.samples

    def test_conversion_gives_correct_prior(self) -> None:
        zeniths = self.samples["zenith"]
        azimuths = self.samples["azimuth"]
        times = self.samples["time"]
        args = zip(*[
            (zenith, azimuth, time, self.ifos)
            for zenith, azimuth, time in zip(zeniths, azimuths, times)
        ])
        ras, decs = zip(*map(bilby.gw.utils.zenith_azimuth_to_ra_dec, *args))
        self.assertGreaterEqual(ks_2samp(self.samples["ra"], ras).pvalue, 0.01)
        self.assertGreaterEqual(ks_2samp(self.samples["dec"], decs).pvalue, 0.01)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
An example of how to use bilby to perform parameter estimation for
non-gravitational wave data. In this case, fitting a linear function to
data with background Gaussian noise. We then compare the result to posteriors
estimated using the Fisher Information Matrix approximation.

"""
import copy

import bilby
import numpy as np

# A few simple setup steps
outdir = "outdir"

np.random.seed(123)


# First, we define our "signal model", in this case a simple linear function
def model(time, m, c):
    return time * m + c


# Now we define the injection parameters which we make simulated data with
injection_parameters = dict(m=0.5, c=0.2)

# For this example, we'll use standard Gaussian noise

# These lines of code generate the fake data. Note the ** just unpacks the
# contents of the injection_parameters when calling the model function.
sampling_frequency = 10
time_duration = 10
time = np.arange(0, time_duration, 1 / sampling_frequency)
N = len(time)
sigma = np.random.normal(1, 0.01, N)
data = model(time, **injection_parameters) + np.random.normal(0, sigma, N)

# Now lets instantiate a version of our GaussianLikelihood, giving it
# the time, data and signal model
likelihood = bilby.likelihood.GaussianLikelihood(time, data, model, sigma)

# From hereon, the syntax is exactly equivalent to other bilby examples
# We make a prior
priors = dict()
priors["m"] = bilby.core.prior.Uniform(0, 5, "m")
priors["c"] = bilby.core.prior.Uniform(-2, 2, "c")
priors = bilby.core.prior.PriorDict(priors)

# And run sampler
result = bilby.run_sampler(
    likelihood=likelihood,
    priors=priors,
    sampler="dynesty",
    nlive=1000,
    injection_parameters=injection_parameters,
    outdir=outdir,
    label="Nested Sampling",
)

# Finally plot a corner plot: all outputs are stored in outdir
result.plot_corner()

fim = bilby.core.fisher.FisherMatrixPosteriorEstimator(likelihood, priors)
result_fim = copy.deepcopy(result)
result_fim.posterior = fim.sample_dataframe("maxL", 10000)
result_fim.label = "Fisher"

bilby.core.result.plot_multiple(
    [result, result_fim], parameters=injection_parameters, truth_color="k"
)
